#include "../config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <mcrypt.h>
#include <mhash.h>
#include <assert.h>

#include "wmpasman.h"
#include "globals.h"
#include "crypto.h"
#include "die.h"

#ifndef va_copy
# define va_copy(d,s) __va_copy(d,s)
#endif

const char *IV_END="end";
const char *IV_EMPTY="empty";

static char *algo=NULL;
static int key_length=0;
static int IV_length=0;

static char *algo2=NULL;
static int key_length2=0;
static int IV_length2=0;

int crypto_set_passphrase(const char *a, const char *p){
    int i;
    MCRYPT foo;
    char *new_algo;
    int new_key_len;
    int new_IV_len;
    KEYGEN keygen;

    assert(a!=NULL);

    if((i=mcrypt_module_self_test((char *)a, NULL))<0){
        set_error("MCrypt self test for %s failed: %s", a, mcrypt_strerror(i));
        return -1;
    }

    if((new_algo=strdup(a))==NULL) return -1;
    new_key_len=mcrypt_module_get_algo_key_size(new_algo, NULL);
    if(new_key_len>MAX_PHRASEHASH_LEN)
        die("Algorithm '%s' has a maximum key size of %d, which is more than the assumed max of %d. Please report this bug!", new_algo, new_key_len, MAX_PHRASEHASH_LEN);

    foo=mcrypt_module_open(new_algo, NULL, (mcrypt_module_is_block_algorithm(new_algo, NULL))?"cfb":"stream", NULL);
    if(foo==MCRYPT_FAILED){
        set_error("MCrypt module open for %s failed", new_algo);
        return -1;
    }
    new_IV_len=mcrypt_enc_get_iv_size(foo);
    mcrypt_module_close(foo);

    if(algo!=NULL){
        memcpy(passphrase_buffer2, passphrase_buffer, MAX_PHRASEHASH_LEN);
        if(algo2!=NULL) free(algo2);
        algo2=algo;
        key_length2=key_length;
        IV_length2=IV_length;
    }

    i=strlen(p);
    keygen.hash_algorithm[0]=MHASH_SHA1;
    keygen.hash_algorithm[1]=MHASH_RIPEMD160;
    keygen.count=i;
    keygen.salt=NULL;
    keygen.salt_size=0;
    if(mhash_keygen_ext(KEYGEN_S2K_SIMPLE, keygen, passphrase_buffer, new_key_len, (unsigned char *)p, i)<0){
        crypto_forget_passphrase();
        set_error("Error hashing passphrase");
        return -2;
    }

    algo=new_algo;
    key_length=new_key_len;
    IV_length=new_IV_len;

    return 0;
}

void crypto_forget_passphrase(){
    if(algo!=NULL && algo!=algo2) free(algo);
    if(algo2!=NULL) free(algo2);
    algo=algo2=NULL;
    memset(passphrase_buffer, 0, MAX_PHRASEHASH_LEN);
    memset(passphrase_buffer2, 0, MAX_PHRASEHASH_LEN);
}

/* Fill IV with a hash based on the salt */
static int hash_salts(char *IV, int length, va_list salts){
    MHASH h;
    int i;
    char *salt, *c, buf[32];
    va_list v;
    int buflen, oldlen;
    hashid type;

    if(length<=0) return 0;

    salt=va_arg(salts, char *);
    if(salt==IV_END){
        return randomize(IV, length);
    }

    if(salt==IV_EMPTY) return 0;

    memset(buf, 'x', 32);
    buflen=32;
    while(length>0){
        oldlen=buflen;
        if(length>=32)     { type=MHASH_HAVAL256; buflen=32; }
        else if(length>=28){ type=MHASH_HAVAL224; buflen=28; }
        else if(length>=24){ type=MHASH_HAVAL192; buflen=24; }
        else if(length>=20){ type=MHASH_HAVAL160; buflen=20; }
        else if(length>=16){ type=MHASH_HAVAL128; buflen=16; }
        else               { type=MHASH_ADLER32;  buflen=4;  }
        assert(mhash_get_block_size(type)==buflen);

        va_copy(v, salts);
        h=mhash_init(type);
        if(h==MHASH_FAILED){
            set_error("Init of hash %s failed", mhash_get_hash_name_static(type));
            return -1;
        }
        if((i=mhash(h, buf, oldlen))){
            set_error("mhash call failed with code %d", i);
            return -1;
        }
        if((i=mhash(h, salt, strlen(salt)))){
            set_error("mhash call failed with code %d", i);
            return -1;
        }
        while((c=va_arg(v, char *))!=IV_END){
            if((i=mhash(h, c, strlen(c)))){
                set_error("mhash call failed with code %d", i);
                return -1;
            }
        }
        mhash_deinit(h, buf);
        memcpy(IV, buf, (buflen<length)?buflen:length);
        IV+=buflen;
        length-=buflen;
        va_end(v);
    }
    return 0;
}

char *crypto_new_IV(int *length, ...){
    char *IV;
    va_list ap;
    
    if(length!=NULL) *length=IV_length;
    if((IV=malloc(IV_length))==NULL) return NULL;

    va_start(ap, length);
    if(hash_salts(IV, IV_length, ap)){
        debug(DEBUG_ERR, "IV creation failed");
        free(IV);
        IV=NULL;
    }
    va_end(ap);
    return IV;
}

static Crypto crypto_new_ptr(const char *IV, const char *a, const char *key, int keylen){
    MCRYPT *m;
    int i;

    assert(a!=NULL);
    assert(IV!=NULL);

    if((m=malloc(sizeof(*m)))==NULL) return NULL;
    
    *m=mcrypt_module_open((char *)a, NULL, (mcrypt_module_is_block_algorithm((char *)a, NULL))?"cfb":"stream", NULL);
    if(*m==MCRYPT_FAILED){
        set_error("MCrypt module open for %s failed", a);
        free(m);
        return NULL;
    }

    if((i=mcrypt_generic_init(*m, (void *)key, keylen, (void *)IV))<0){
        mcrypt_module_close(*m);
        free(m);
        set_error("MCrypt generic init failed: %s", mcrypt_strerror(i));
        return NULL;
    }

    return m;
}

Crypto crypto_new(const char *IV){
    return crypto_new_ptr(IV, algo, passphrase_buffer, key_length);
}

void crypto_free(Crypto c){
    mcrypt_generic_deinit(*(MCRYPT *)c);
    mcrypt_module_close(*(MCRYPT *)c);
    free(c);
}

int crypto_encrypt(Crypto c, int *len, char *string){
    int l;

    l=strlen(string);
    if(len!=NULL) *len=l;
    return crypto_encrypt_binary(c, l, string);
}

int crypto_encrypt_binary(Crypto c, int l, char *string){
    int i;

    if((i=mcrypt_generic(*(MCRYPT *)c, string, l))<0)
        set_error("MCrypt encryption failed: %s", mcrypt_strerror(i));
    return i;
}

int crypto_decrypt(Crypto c, int l, char *string){
    int i;

    if((i=mdecrypt_generic(*(MCRYPT *)c, string, l))<0)
        set_error("MCrypt decryption failed: %s", mcrypt_strerror(i));
    return i;
}

static int crypto_recrypt(int len, char *string, const char *new_IV, const char *old_IV){
    Crypto c;

    assert(len<MAX_PASSWORD_LEN);
    assert(algo2!=NULL);

    if(string!=password_buffer) memcpy(password_buffer, string, len);

    if((c=crypto_new_ptr(old_IV, algo2, passphrase_buffer2, key_length2))==NULL) return -1;
    crypto_decrypt(c, len, password_buffer);
    crypto_free(c);
    password_buffer[MAX_PASSWORD_LEN-1]='\0';
    if((c=crypto_new_ptr(new_IV, algo, passphrase_buffer, key_length))==NULL) return -1;
    crypto_encrypt(c, NULL, password_buffer);
    crypto_free(c);

    if(string!=password_buffer) memcpy(string, password_buffer, len);

    return 0;
}

const char *crypto_recrypt_salted(int len, char *string, int *new_IV_length, ...){
    char *old_IV, *new_IV;
    va_list ap, ap2;

    assert(algo2!=NULL);

    if((old_IV=malloc(IV_length2))==NULL) return NULL;
    if((new_IV=malloc(IV_length))==NULL){
        free(old_IV);
        return NULL;
    }
    if(new_IV_length!=NULL) *new_IV_length=IV_length;

    va_start(ap, new_IV_length);
    va_copy(ap2, ap);
    if(hash_salts(old_IV, IV_length2, ap2)){
        debug(DEBUG_ERR, "IV creation failed");
        free(old_IV);
        free(new_IV);
        return NULL;
    }
    if(hash_salts(new_IV, IV_length, ap)){
        debug(DEBUG_ERR, "IV creation failed");
        free(old_IV);
        free(new_IV);
        return NULL;
    }
    va_end(ap2);
    va_end(ap);

    if(crypto_recrypt(len, string, new_IV, old_IV)){
        free(old_IV);
        free(new_IV);
        return NULL;
    }
    free(old_IV);
    return new_IV;
}

const char *crypto_recrypt_IV(int len, char *string, int *new_IV_length, const char *old_IV){
    char *new_IV;

    assert(algo2!=NULL);

    if((new_IV=malloc(IV_length))==NULL) return NULL;
    if(new_IV_length!=NULL) *new_IV_length=IV_length;

    if(randomize(new_IV, IV_length)){
        debug(DEBUG_ERR, "IV creation failed");
        free(new_IV);
        return NULL;
    }

    if(crypto_recrypt(len, string, new_IV, old_IV)){
        free(new_IV);
        return NULL;
    }
    return new_IV;
}

int randomize(char *IV, int length){
    static FILE *fp=NULL;

    if(length==0) return 0;
    if(fp==NULL){
        if((fp=fopen("/dev/urandom", "rb"))==NULL) return -1;
    }
    fread(IV, sizeof(*IV), length, fp);
    return 0;
}
