#ifndef CRYPTO_H
#define CRYPTO_H

/* Set the passphrase used for all new cryptos. Returns -1 if the passphrase
 * couldn't be set, -2 if the error was so fatal that crypto_forget_passphrase
 * was called. */
int crypto_set_passphrase(const char *algo, const char *passphrase);

/* Clear the passphrase memorized above */
void crypto_forget_passphrase();

typedef void *Crypto;

extern const char *IV_END;
extern const char *IV_EMPTY;

/* Get a new IV. On error, NULL will be returned. Otherwise, an IV buffer will
 * be returned, which may be passed to crypto_new. 'length' will be set to the
 * length of the IV buffer if it is non-NULL. The ellipses are either IV_EMPTY,
 * or a series of zero or more salt strings with IV_END ending the list. If
 * IV_EMPTY is specified, the IV buffer will be filled with memory garbage. If
 * salt strings are provided, the IV buffer will be filled with binary data
 * that depends on those strings. Otherwise, completely random IV buffer will
 * be returned (otherwise the buffer depends on the salt). The IV buffer may be
 * freed with free().
 *
 * Note that the IV returned depends on the values of the latest call to
 * crypto_set_passphrase! */
char *crypto_new_IV(int *length, ...);

/* Create a new Crypto.
 *
 * Note that the following may fail:
 *  void foo(char *in){
 *      char *out=strdup(in);
 *      Crypto c=crypto_new(NULL);
 *   
 *      crypto_encrypt(c, out);
 *      crypto_decrypt(c, out);
 *      assert(in==out);  // <= NO!
 *  }
 *
 *  However, this should always work fine:
 *  void bar(char *in){
 *      char *out=strdup(in);
 *      const char *IV=crypto_new_IV(NULL, END_STRING);
 *      Crypto c1=crypto_new(IV);
 *      Crypto c2=crypto_new(IV);
 *   
 *      crypto_encrypt(c1, out);
 *      crypto_decrypt(c2, out);
 *      assert(in==out);  // <= YES!
 *  }
 */
Crypto crypto_new(const char *IV);

/* Free a Crypto */
void crypto_free(Crypto c);

/* Encrypt the given string in place. 0 on success, -1 on error.
 * The length of the string is written into len (if non-NULL) as a convenience,
 * as the encrypted string may contain nulls. */
int crypto_encrypt(Crypto c, int *len, char *string);

/* Similar to the above, but you supply the length so binary data can be
 * encrypted. */
int crypto_encrypt_binary(Crypto c, int l, char *string);

/* Decrypt the given string in place. 0 on success, -1 on error. 
 * The length parameter is necessary as the encrypted string may contain
 * nulls. */
int crypto_decrypt(Crypto c, int len, char *string);

/* After crypto_set_passphrase() has been called for the second time, this
 * function can be used to decrypt 'string' using the old passphrase and then
 * encrypt it with the new one.
 *
 * crypto_recrypt_salted will generate the old and new IVs based on the elipses
 * salts as with crypto_new_IV. crypto_recrypt_IV will decrypt using the given
 * IV, and will generate a new random IV for encryption. In either case, the
 * new IV will be returned, or NULL on error. */
const char *crypto_recrypt_salted(int len, char *string, int *new_IV_length, ...);
const char *crypto_recrypt_IV(int len, char *string, int *new_IV_length, const char *old_IV);

/* Fill a buffer with random data */
int randomize(char *buf, int length);

#endif
