#include "config.h"

#include <string.h>
#include <ctype.h>
#include <gtk/gtk.h>

/* We treat password_edit_buffer as a "gapped buffer", so the common case of
 * "pick an insertion point and type a bunch" is fast. The basic idea is that
 * the beginning of the buffer contains the text up to the insertion point, and
 * the text after the insertion point is at the very end of the buffer.
 *
 * This function moves the insertion point, by moving characters from one end
 * of the buffer to the other. Note that if we call move_gap(MAX_PASSWORD_LEN),
 * all the text ends up at the start of the buffer and we can treat it like a
 * normal C string ;)
 */
static void move_gap(GtkTextBuffer *b, int idx){
    char *pos;
    char *buffer, *gap_pos;
    int buflen, gap_len;

    buffer=(char *)g_object_get_data(G_OBJECT(b), "buffer");
    buflen=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "buflen"));
    gap_pos=(char *)g_object_get_data(G_OBJECT(b), "gap_pos");
    gap_len=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "gap_len"));

    /* If idx is past the end of the string, put it at the end. */
    if(idx>buflen-1-gap_len) idx=buflen-1-gap_len;

    /* Move the gap, or abort if it's not to be moved */
    if((pos=buffer+idx)==gap_pos) return;
    if(gap_len>0){
        if(pos<gap_pos){
            memmove(pos+gap_len, pos, gap_pos-pos);
        } else {
            memmove(gap_pos, gap_pos+gap_len, pos-gap_pos);
        }
    }
    gap_pos=pos;

    /* Be nice, zero out the gap */
    if(gap_len>0) memset(gap_pos, 0, gap_len);

    g_object_set_data(G_OBJECT(b), "buffer", (gpointer)buffer);
    g_object_set_data(G_OBJECT(b), "buflen", GINT_TO_POINTER(buflen));
    g_object_set_data(G_OBJECT(b), "gap_pos", (gpointer)gap_pos);
    g_object_set_data(G_OBJECT(b), "gap_len", GINT_TO_POINTER(gap_len));
}

/* Signal handler for inserting text. So it also handles inserting text into
 * the gapped buffer. */
static void insert_text(GtkTextBuffer *b, GtkTextIter *pos, gchar *text, gint len, gpointer data){
    int i;
    char *gap_pos;
    int gap_len;
    gboolean visible;
    /* XXX: Why does gcc think 'stars' could be used uninitialized? */
    GString *stars;
    
    /* Move the insertion point */
    move_gap(b, gtk_text_iter_get_offset(pos));

    gap_pos=(char *)g_object_get_data(G_OBJECT(b), "gap_pos");
    gap_len=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "gap_len"));
    visible=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "visible"));

    /* If not visible, create the array of stars */
    if(!visible) stars=g_string_sized_new(len);

    /* Insert the text (or beep if there's no room) and shrink the gap, and
     * simultaneously build up a string of either the text or *s for insertion
     * into the GtkTextBuffer. */
    for(i=0; i<len; i++){
        if(gap_len<=0){
            gdk_beep();
            break;
        }
        *(gap_pos++)=text[i];
        if(!visible) g_string_append_c(stars, '*');
        gap_len--;
    }

    /* Insert the string (either text or *s) into the GtkTextBuffer. We have to
     * disable the current callback first to prevent recursion. Then, don't
     * allow the other signal handlers to fire on the original event (since we
     * already sent them a transformed one). */
    g_signal_handlers_block_by_func(G_OBJECT(b), G_CALLBACK(insert_text), data);
    gtk_text_buffer_insert(b, pos, visible?text:stars->str, i);
    if(!visible) g_string_free(stars, TRUE);
    g_signal_handlers_unblock_by_func(G_OBJECT(b), G_CALLBACK(insert_text), data);
    g_signal_stop_emission_by_name(G_OBJECT(b), "insert-text");

    g_object_set_data(G_OBJECT(b), "gap_pos", (gpointer)gap_pos);
    g_object_set_data(G_OBJECT(b), "gap_len", GINT_TO_POINTER(gap_len));
}

/* Signal handler for deleting text. So it also handles removing text from the
 * gapped buffer. */
static void delete_range(GtkTextBuffer *b, GtkTextIter *s, GtkTextIter *e, gpointer data){
    char *gap_pos;
    int gap_len;

    /* Move insertion point to the beginning of the deleted text (so, the
     * to-be-deleted text is actually after the gap) */
    move_gap(b, gtk_text_iter_get_offset(s));

    /* Enlarge the gap, and zero out that to-be-deleted text */
    gap_pos=(char *)g_object_get_data(G_OBJECT(b), "gap_pos");
    gap_len=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "gap_len"));
    gap_len+=gtk_text_iter_get_offset(e)-gtk_text_iter_get_offset(s);
    memset(gap_pos, 0, gap_len);
    g_object_set_data(G_OBJECT(b), "gap_len", GINT_TO_POINTER(gap_len));

    /* We let the event propagate unchanged, since deleting *s or text works
     * the same either way. */
}

/* toggle the visibility of the GtkTextBuffer */
void my_text_buffer_set_visibility(GtkTextBuffer *b, gboolean visible){
    int i, inspos, selpos, buflen;
    char *buffer;
    GtkTextIter iter;
    GtkTextMark *ins, *sel;

    buffer=(char *)g_object_get_data(G_OBJECT(b), "buffer");
    buflen=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "buflen"));
    
    /* We'll be inserting and deleting, but we don't want to mess with
     * password_edit_buffer */
    g_signal_handlers_block_by_func(G_OBJECT(b), G_CALLBACK(insert_text), NULL);
    g_signal_handlers_block_by_func(G_OBJECT(b), G_CALLBACK(delete_range), NULL);

    /* Be nice and save the current cursor position and selection bounds */
    ins=gtk_text_buffer_get_insert(b);
    sel=gtk_text_buffer_get_selection_bound(b);
    gtk_text_buffer_get_iter_at_mark(b, &iter, ins);
    inspos=gtk_text_iter_get_offset(&iter);
    gtk_text_buffer_get_iter_at_mark(b, &iter, sel);
    selpos=gtk_text_iter_get_offset(&iter);
    
    /* Erase the buffer, and re-fill it with either text or *s */
    move_gap(b, buflen);
    gtk_text_buffer_set_text(b, "", 0);
    if(visible){
        gtk_text_buffer_insert_at_cursor(b, buffer, -1);
    } else {
        for(i=0; i<buflen; i++){
            if(!buffer[i]) break;
            gtk_text_buffer_insert_at_cursor(b, "*", 1);
        }
    }

    /* Restore the old cursor position and selection */
    gtk_text_buffer_get_iter_at_offset(b, &iter, inspos);
    gtk_text_buffer_place_cursor(b, &iter);
    if(inspos!=selpos){
        gtk_text_buffer_get_iter_at_offset(b, &iter, selpos);
        gtk_text_buffer_move_mark(b, sel, &iter);
    }
    
    /* And finally, re-enable the insert and delete signal handlers */
    g_signal_handlers_unblock_by_func(G_OBJECT(b), G_CALLBACK(insert_text), NULL);
    g_signal_handlers_unblock_by_func(G_OBJECT(b), G_CALLBACK(delete_range), NULL);

    g_object_set_data(G_OBJECT(b), "visible", GINT_TO_POINTER(visible));
}

GtkTextBuffer *my_text_buffer_new(char *buffer, int buflen){
    GtkTextBuffer *b;

    memset(buffer, 0, buflen);

    b=gtk_text_buffer_new(NULL);
    g_object_set_data(G_OBJECT(b), "buffer", (gpointer)buffer);
    g_object_set_data(G_OBJECT(b), "buflen", GINT_TO_POINTER(buflen));
    g_object_set_data(G_OBJECT(b), "gap_pos", (gpointer)buffer);
    g_object_set_data(G_OBJECT(b), "gap_len", GINT_TO_POINTER(buflen-1));
    g_object_set_data(G_OBJECT(b), "visible", GINT_TO_POINTER(FALSE));
    g_signal_connect(G_OBJECT(b), "insert-text", G_CALLBACK(insert_text), NULL);
    g_signal_connect(G_OBJECT(b), "delete-range", G_CALLBACK(delete_range), NULL);

    return b;
}

char *my_text_buffer_get_buffer(GtkTextBuffer *b){
    move_gap(b, GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "buflen")));
    return (char *)g_object_get_data(G_OBJECT(b), "buffer");
}

gboolean my_text_buffer_get_visibility(GtkTextBuffer *b){
    return GPOINTER_TO_INT(g_object_get_data(G_OBJECT(b), "visible"));
}
