#ifndef PASSWORD_H
#define PASSWORD_H

#define LINE_CHARS 9

typedef struct _Password *Password;

/* Create a new password object. Note that the string 'pass' will be
 * overwritten. */
Password password_new(const char *line1, const char *line2, char *pass);

/* Free a password object */
void password_free(Password p);

/* Get the labels for a password object. Note that the labels may be
 * silently truncated. */
void password_get_labels(Password p, const char **line1, const char **line2);

/* Get the password in a password object. Note that the value returned by
 * password_get_pass is a pointer to statically allocated memory, so don't try
 * to free it and don't assume it will remain unaltered over multiple calls to
 * any of the password_* functions. */
const char *password_get_pass(Password p);

/* If the passphrase is changed, use this to update the password entry */
int password_recrypt_pass(Password pw);


/**********************************************/

typedef struct {
    int num;       /* Number of passwords in the list, you'll need this */
    int capacity;  /* Capacity, you can ignore this */
    Password *p;   /* The actual list */
} *PWList;

/* Load saved passwords from a file. Returns the number of passwords read, -1
 * on error. If passwords were read, list will point to an array containing the
 * passwords read. Caller is responsible for freeing the individual Password
 * structs and the list itself. */
int pwlist_load(PWList list, const char *filename);

/* Save passwords to a file. Returns 0 on success, -1 on error. */
int pwlist_save(PWList list, const char *filename);

/* Create a new PWList. capacity is the number of entries to allocate space for
 * before relloc is needed. */
PWList pwlist_new(int capacity);

/* Free a PWList. pwlist_free_all also frees all the Passwords in the list */
void pwlist_free(PWList p);
void pwlist_free_all(PWList p);

/* Return the number of Passwords in the list */
int pwlist_count(PWList p);

/* Add an entry. Returns 0 on success, -1 on error. */
int pwlist_push(PWList p, Password pw);

/* Find an entry by label, -1 or NULL if not found */
int pwlist_find_i(PWList p, const char *line1, const char *line2);
Password pwlist_find_p(PWList p, const char *line1, const char *line2);

/* Get an entry. pwlist_get_i returns the Password at the index, pwlist_get_p
 * returns the index of the password. NULL/-1 on error. */
Password pwlist_get_i(PWList p, int idx);
int pwlist_get_p(PWList p, Password pw);

/* Remove an entry. pwlist_remove_i returns the Password that was at the index,
 * and pwlist_remove_p returns the index the password was at. */
Password pwlist_remove_i(PWList p, int idx);
int pwlist_remove_p(PWList p, Password pw);

/* Remove a Password from the list and free it. Returns 0 on success. */
int pwlist_free_i(PWList p, int idx);
int pwlist_free_p(PWList p, Password pw);

/* If the passphrase is changed, use this to update the password entries */
int pwlist_recrypt_passes(PWList p);

#endif
